/*
**      cdecl -- C gibberish translator
**      src/c_lang.h
**
**      Copyright (C) 2017-2020  Paul J. Lucas, et al.
**
**      This program is free software: you can redistribute it and/or modify
**      it under the terms of the GNU General Public License as published by
**      the Free Software Foundation, either version 3 of the License, or
**      (at your option) any later version.
**
**      This program is distributed in the hope that it will be useful,
**      but WITHOUT ANY WARRANTY; without even the implied warranty of
**      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**      GNU General Public License for more details.
**
**      You should have received a copy of the GNU General Public License
**      along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef cdecl_c_lang_H
#define cdecl_c_lang_H

/**
 * @file
 * Declares constants, types, and functions for C/C++ language versions.
 */

// local
#include "cdecl.h"                      /* must go first */
#include "typedefs.h"

/// @cond DOXYGEN_IGNORE

// standard
#include <stdbool.h>

_GL_INLINE_HEADER_BEGIN
#ifndef C_LANG_INLINE
# define C_LANG_INLINE _GL_INLINE
#endif /* C_LANG_INLINE */

/// @endcond

/**
 * @defgroup c-lang-group C/C++ Language Versions
 * Constants, types, and functions for C/C++ language versions.
 * @{
 */

///////////////////////////////////////////////////////////////////////////////

// languages supported
#define LANG_NONE     (0u)              /**< No languages. */
#define LANG_ALL      (~LANG_NONE)      /**< All supported languages. */

#define LANG_C_OLD    LANG_C_KNR        /**< Oldest supported C language. */
#define LANG_C_KNR    (1u << 0)         /**< K&R (pre-ANSI) C. */
#define LANG_C_89     (1u << 1)         /**< C 89 (first ANSI C). */
#define LANG_C_95     (1u << 2)         /**< C 95. */
#define LANG_C_99     (1u << 3)         /**< C 99. */
#define LANG_C_11     (1u << 4)         /**< C 11. */
#define LANG_C_18     (1u << 5)         /**< C 18. */
#define LANG_C_NEW    LANG_C_18         /**< Newest supported C language. */
#define LANG_C_ALL    LANG_MAX(C_NEW)   /**< All C languages. */

#define LANG_CPP_OLD  LANG_CPP_98       /**< Oldest supported C++ language. */
#define LANG_CPP_98   (1u << 8)         /**< C++ 98. */
#define LANG_CPP_03   (1u << 9)         /**< C++ 03. */
#define LANG_CPP_11   (1u << 10)        /**< C++ 11. */
#define LANG_CPP_14   (1u << 11)        /**< C++ 14. */
#define LANG_CPP_17   (1u << 12)        /**< C++ 17. */
#define LANG_CPP_20   (1u << 13)        /**< C++ 20. */
#define LANG_CPP_NEW  LANG_CPP_20       /**< Newest supported C++ language. */
#define LANG_CPP_ALL  LANG_MIN(CPP_OLD) /**< All C++ languages. */

// bit masks
#define LANG_MASK_C   0x00FFu           /**< C languages bitmask. */
#define LANG_MASK_CPP 0xFF00u           /**< C++ languages bitmask. */

/** Maximum allowed language, C & C++. */
#define LANG_MAX(L)               (LANG_ ## L | (LANG_ ## L - 1u))

/** Minimum allowed language, C & C++. */
#define LANG_MIN(L)               (~(LANG_ ## L - 1u))

/** Maximum allowed language, C only. */
#define LANG_C_MAX(L)             LANG_MAX( C_ ## L )

/** Minimum allowed language, C only. */
#define LANG_C_MIN(L)             (LANG_MIN( C_ ## L ) & LANG_MASK_C)

/* Maximum allowed language, C++ only. */
#define LANG_CPP_MAX(L)           (LANG_MAX( CPP_ ## L ) & LANG_MASK_CPP)

/* Minimum allowed language, C++ only. */
#define LANG_CPP_MIN(L)           LANG_MIN( CPP_ ## L )

/* Maximum allowed language, C & C++ seperately. */
#define LANG_C_CPP_MAX(CL,CPPL)   (LANG_C_MAX(CL) | LANG_CPP_MAX(CPPL))

/* Minimum allowed language, C & C++ seperately. */
#define LANG_C_CPP_MIN(CL,CPPL)   (LANG_C_MIN(CL) | LANG_CPP_MIN(CPPL))

/**
 * C/C++ language(s)/literal pairs: for the given language(s) only, use the
 * given literal.  This allows different languages to use different literals,
 * e.g., "_Noreturn" for C and "noreturn" for C++.
 */
struct c_lang_lit {
  c_lang_id_t   lang_ids;               ///< Language(s) literal is in.
  char const   *literal;                ///< The literal.
};

////////// extern functions ///////////////////////////////////////////////////

/**
 * Gets the <code>\ref c_lang_id_t</code> corresponding to the given string
 * (case insensitive).
 *
 * @param name The language name to get the corresponding <code>\ref
 * c_lang_id_t</code> for.
 * @return Returns said language or <code>\ref LANG_NONE</code> if \a name
 * doesn't correspond to any supported language.
 */
C_WARN_UNUSED_RESULT
c_lang_id_t c_lang_find( char const *name );

/**
 * Gets whether \a lang_id is any version of C.
 *
 * @param lang_id The bitwise-or of language(s) to check.
 * @return Returns `true` only if \a lang_id is C.
 *
 * @sa c_lang_is_cpp()
 */
C_LANG_INLINE C_WARN_UNUSED_RESULT
bool c_lang_is_c( c_lang_id_t lang_id ) {
  return (lang_id & LANG_MASK_C) != LANG_NONE;
}

/**
 * Gets whether \a lang_id is any version of C++.
 *
 * @param lang_id The bitwise-or of language(s) to check.
 * @return Returns `true` only if \a lang_id is C++.
 *
 * @sa c_lang_is_c()
 */
C_LANG_INLINE C_WARN_UNUSED_RESULT
bool c_lang_is_cpp( c_lang_id_t lang_id ) {
  return (lang_id & LANG_MASK_CPP) != LANG_NONE;
}

/**
 * Shorthand for the common case of getting whether the current language is any
 * version of C.
 *
 * @return Returns `true` only if the current language is C.
 */
#define C_LANG_IS_C()             c_lang_is_c( opt_lang )

/**
 * Shorthand for the common case of getting whether the current language is any
 * version of C++.
 *
 * @return Returns `true` only if the current language is C++.
 */
#define C_LANG_IS_CPP()           c_lang_is_cpp( opt_lang )

/**
 * Gets the literal appropriate for the current language.
 *
 * @param lang_lit A c_lang_lit_t array.  The last element _must_ always have a
 * `lang_ids` value of #LANG_ALL.
 * @return Returns said literal.
 */
C_WARN_UNUSED_RESULT
char const* c_lang_literal( c_lang_lit_t const lang_lit[] );

/**
 * Gets a comma-separated string of all supported language names.
 *
 * @return Returns said string.
 */
C_WARN_UNUSED_RESULT
char const* c_lang_names( void );

/**
 * Gets the oldest language of the given set.
 *
 * @param lang_id The bitwise-or of language(s).
 * @return Returns said language.
 */
C_LANG_INLINE C_WARN_UNUSED_RESULT
c_lang_id_t c_lang_oldest( c_lang_id_t lang_id ) {
  return lang_id & ~(lang_id - 1u);
}

/**
 * Gets the printable name of \a lang_id.
 *
 * @param lang_id The <code>\ref c_lang_id_t</code> to get the name of.
 * @return Returns said name.
 */
C_WARN_UNUSED_RESULT
char const* c_lang_name( c_lang_id_t lang_id );

/**
 * Shorthand for the common case of getting the name of the current language.
 *
 * @return Returns said name.
 */
#define C_LANG_NAME()             c_lang_name( opt_lang )

/**
 * Sets the current language and the corresponding prompt.
 *
 * @param lang_id The language to set.
 */
void c_lang_set( c_lang_id_t lang_id );

///////////////////////////////////////////////////////////////////////////////

/** @} */

_GL_INLINE_HEADER_END

#endif /* cdecl_c_lang_H */
/* vim:set et sw=2 ts=2: */
